/* $Id: format.c,v 1.5 2008/07/18 13:32:17 nanardon Exp $ */
#include <postgres.h>
#include <fmgr.h>

#include "format.h"
#include "rpmver.h"
#include "pgrpm.h"

PG_FUNCTION_INFO_V1(rpmSenseFlagtoText);
Datum
rpmSenseFlagtoText(PG_FUNCTION_ARGS) {
    int fl = PG_GETARG_INT32(0);
    char * res = palloc(4);
    char * ptr = res;
    text * tres;
    
    if(fl & RPMSENSE_LESS) *ptr++ = '<';
    if(fl & RPMSENSE_GREATER) *ptr++ = '>';
    if(fl & RPMSENSE_EQUAL) *ptr++ = '=';
    *ptr = '\0';

    tres = (text*) palloc(VARHDRSZ+strlen(res));
#ifdef SET_VARSIZE
    SET_VARSIZE(tres, strlen(res)+VARHDRSZ);
#else
    VARATT_SIZEP(tres) = strlen(res)+VARHDRSZ;
#endif
    memcpy(VARDATA(tres), res, strlen(res));
    
    PG_RETURN_TEXT_P(tres);
}

PG_FUNCTION_INFO_V1(rpmSenseTexttoFlag);
Datum
rpmSenseTexttoFlag(PG_FUNCTION_ARGS) {
    text * tflag = PG_GETARG_TEXT_P(0);
    char * ptr;
    int flag = 0;

    for(ptr = VARDATA(tflag); ptr <= VARDATA(tflag) + VARSIZE(tflag); ptr++) {
        switch(*ptr) {
            case '<':
                flag |= RPMSENSE_LESS;
            break;
            case '>':
                flag |= RPMSENSE_GREATER;
            break;
            case '=':
                flag |= RPMSENSE_EQUAL;
            break;
            default:
            break;
        }
    }

    PG_RETURN_INT32(flag);
}

PG_FUNCTION_INFO_V1(rpmfflagstoText);
Datum
rpmfflagstoText(PG_FUNCTION_ARGS) {
    int fl = PG_GETARG_INT32(0);
    text * tresult = palloc(VARHDRSZ+ 11);
    char *ptr = VARDATA(tresult);
#ifdef SET_VARSIZE
    SET_VARSIZE(tresult, VARHDRSZ + 11);
#else
    VARATT_SIZEP(tresult) = VARHDRSZ + 11;
#endif

    *ptr++ = fl & RPMFILE_CONFIG    ? 'C' : '.';
    *ptr++ = fl & RPMFILE_DOC       ? 'D' : '.';
    *ptr++ = fl & RPMFILE_ICON      ? 'I' : '.';
    *ptr++ = fl & RPMFILE_MISSINGOK ? 'M' : '.';
    *ptr++ = fl & RPMFILE_NOREPLACE ? 'N' : '.';
    *ptr++ = fl & RPMFILE_SPECFILE  ? 'S' : '.';
    *ptr++ = fl & RPMFILE_GHOST     ? 'G' : '.';
    *ptr++ = fl & RPMFILE_LICENSE   ? 'L' : '.';
    *ptr++ = fl & RPMFILE_README    ? 'R' : '.';
    *ptr++ = fl & RPMFILE_UNPATCHED ? 'U' : '.';
    *ptr++ = fl & RPMFILE_PUBKEY    ? 'P' : '.';
    
    PG_RETURN_TEXT_P(tresult);
}

PG_FUNCTION_INFO_V1(rpmfmodetoText);
Datum
rpmfmodetoText(PG_FUNCTION_ARGS) {
    int mode = PG_GETARG_INT32(0);
    text * tresult = palloc(VARHDRSZ+ 10);
    char *ptr = VARDATA(tresult);
#ifdef SET_VARSIZE
    SET_VARSIZE(tresult, VARHDRSZ + 10);
#else
    VARATT_SIZEP(tresult) = VARHDRSZ + 10;
#endif

    *ptr++ = 
        S_ISLNK(mode)  ? 'l' :
        S_ISFIFO(mode) ? 'f' :
        S_ISDIR(mode)  ? 'd' :
        S_ISCHR(mode)  ? 'c' :
        S_ISBLK(mode)  ? 'b' :
        '-';
    *ptr++ = mode & S_IRUSR ? 'r' : '-';
    *ptr++ = mode & S_IWUSR ? 'w' : '-';
    *ptr++ = mode & S_ISUID ? 's' : mode & S_IXUSR ? 'x' : '-';
    
    *ptr++ = mode & S_IRGRP ? 'r' : '-';
    *ptr++ = mode & S_IWGRP ? 'w' : '-';
    *ptr++ = mode & S_ISGID ? 's' : mode & S_IXGRP ? 'x' : '-';
    
    *ptr++ = mode & S_IROTH ? 'r' : '-';
    *ptr++ = mode & S_IWOTH ? 'w' : '-';
    *ptr++ = mode & S_ISVTX ? 't' :mode & S_IXOTH ? 'x' : '-';

    
    PG_RETURN_TEXT_P(tresult);
}
