/* $Id: pgheader.c,v 1.31 2008/07/24 06:51:10 nanardon Exp $ */
#include <postgres.h>
#include <fmgr.h>
#include <utils/builtins.h>
#include <funcapi.h>

#include "pgrpm.h"
#include "pgrpmlib.h"
#include "pgheader.h"

#if RPMMAJOR == 5
static int headerGetEntry(Header h, int_32 tag, hTYP_t type, void ** p, hCNT_t c)
{
    HE_t he = calloc(1, sizeof(*he));
    int rc;
	
    /* Always ensure to initialize */
    *(void **)p = NULL;
    he->tag = (rpmTag)tag;
    rc = headerGet(h, he, 0);
    if (rc) {
	if (type) *type = he->t;
	if (p) *(void **) p = he->p.ptr;
	if (c) *c = he->c;
    }
    if (he) free(he);

    return rc;
}

static void * headerFreeData(const void * data, rpmTagType type)
{
    if (data)
	free((void *)data);
    return NULL;
}
#endif

bytea *
headergetDatum(Header h)
{
    bytea * out;
#if RPMMAJOR == 5
    size_t size = 0;
    void * ptr = headerUnload(h, &size);
#else
    void * ptr = headerUnload(h);
    int size = headerSizeof(h, HEADER_MAGIC_NO);
#endif
    out = palloc(size + VARHDRSZ);
#ifdef SET_VARSIZE
    SET_VARSIZE(out, size + VARHDRSZ);
#else
    VARATT_SIZEP(out) = size + VARHDRSZ;
#endif
    memcpy(VARDATA(out), ptr, size);
    _free(ptr);
    return out;
}

Header
Datumgetheader(bytea * in, int * size)
{
    Header h;
    h = headerLoad(VARDATA(in));
    return h;
}
    
PG_FUNCTION_INFO_V1(header_bytea_in);

Datum
header_bytea_in(PG_FUNCTION_ARGS)
{
    Datum dest;
    rpmts ts = NULL;
    const char * msg = NULL;
    Header h = NULL;
    bytea * out = NULL;
    unsigned int header_size = 0;

    char tempfile[255];
    int temp;
    rpmRC rc;
    FD_t fd;

    dest = PG_GETARG_DATUM(0);
    header_size = VARSIZE(dest) - VARHDRSZ;

    strcpy(tempfile, "/tmp/XXXXXX");
    if((temp = mkstemp(tempfile)) < 0) {
            ereport(ERROR,
                (errcode(ERRCODE_INSUFFICIENT_RESOURCES),
                errmsg("Problem creating temp file")));
    }
    unlink(tempfile);
    if (write(temp, VARDATA(dest), header_size) != header_size) {
        /* catch write error */
        ereport(ERROR,
                (errcode(ERRCODE_INSUFFICIENT_RESOURCES),
                 errmsg("Can't write to tempfile %s", tempfile)));
    }
    lseek(temp, 0L, SEEK_SET);
    ts = rpmtsCreate();
    rpmtsSetVSFlags(ts, _RPMVSF_NOSIGNATURES);

    fd = fdDup(temp);
    rc = rpmReadHeader(ts, fd, &h, &msg);
    close(temp);
    fdClose(fd);
    if (rc != RPMRC_OK) {
            ereport(ERROR,
                (errcode(ERRCODE_INVALID_BINARY_REPRESENTATION),
                errmsg("Problem while reading header, rpm error: %d", rc)));
    }

    /* Sanity Check */
    
    out = headergetDatum(h);
    h = headerFree(h);
    ts = rpmtsFree(ts);
    PG_RETURN_BYTEA_P(out);
}

PG_FUNCTION_INFO_V1(header_in);

Datum
header_in(PG_FUNCTION_ARGS)
{
    Datum dest;
    rpmts ts = rpmtsCreate();
    const char * msg = NULL;
    
    rpmtsSetVSFlags(ts, _RPMVSF_NOSIGNATURES);
    
    dest = DirectFunctionCall1(byteain, CStringGetDatum(PG_GETARG_CSTRING(0)));
    if (headerCheck(ts, VARDATA(dest), VARSIZE(dest)-VARHDRSZ, &msg) != RPMRC_OK) {
        ereport(ERROR,
            (errcode(ERRCODE_INVALID_BINARY_REPRESENTATION),
            errmsg("Problem while reading header: %s", msg)));
    }
    
    ts = rpmtsFree(ts);
    _free(msg);
    PG_RETURN_BYTEA_P(dest);
}

PG_FUNCTION_INFO_V1(header_out);

Datum
header_out(PG_FUNCTION_ARGS)
{
    PG_RETURN_CSTRING(DirectFunctionCall1(byteaout, PG_GETARG_DATUM(0)));
} 

PG_FUNCTION_INFO_V1(headerfromfile);

Datum
headerfromfile(PG_FUNCTION_ARGS)
{
    rpmts ts = rpmtsCreate();
    char * filename;
    FD_t fd;
    Header h = NULL;
    rpmRC rc = 0;
    bytea * out = NULL;
    
    filename = GET_STR(PG_GETARG_TEXT_P(0));
    if ((fd = Fopen(filename, "r"))) {
        rpmtsSetVSFlags(ts, _RPMVSF_NOPAYLOAD | _RPMVSF_NOSIGNATURES);
        
        rc = rpmReadPackageFile(ts, fd, filename, &h);
        if (rc == RPMRC_OK) {
            out = headergetDatum(h);
        } else {
            ereport(ERROR,
                (errcode(ERRCODE_INVALID_BINARY_REPRESENTATION),
                errmsg("Can't get valid header from file, rpm error: %d", rc)));
        }
    } else {
        ereport(ERROR,
            (errcode(ERRCODE_INSUFFICIENT_RESOURCES),
             errmsg("Can't open file")));
    }
    Fclose(fd);
    ts = rpmtsFree(ts);
    h = headerFree(h);
    PG_RETURN_BYTEA_P(out);
}

PG_FUNCTION_INFO_V1(header_queryformat);

Datum
header_queryformat(PG_FUNCTION_ARGS)
{
    Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
    char * format = GET_STR(PG_GETARG_TEXT_P(1));
    char * res = NULL;
    text * tres = NULL;
    int size;
    res = headerSprintf(h, format, rpmTagTable, rpmHeaderFormats, NULL);
    if (res) {
        size = strlen(res) + VARHDRSZ;
        tres = palloc(size);
#ifdef SET_VARSIZE
        SET_VARSIZE(tres, size);
#else
        VARATT_SIZEP(tres) = size;
#endif
        memcpy(VARDATA(tres), res, size - VARHDRSZ);
    }    
    _free(res); /* this malloc come from rpmlib */
    h = headerFree(h);
    if (tres)
        PG_RETURN_TEXT_P(tres);
    else
        PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(header_hastag);

Datum
header_hastag(PG_FUNCTION_ARGS)
{
    Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
    int tag = PG_GETARG_INT32(1);
    int res;
    
    res = headerIsEntry(h, tag);
    headerFree(h);

    PG_RETURN_BOOL(res);
}

PG_FUNCTION_INFO_V1(header_size);

Datum
header_size(PG_FUNCTION_ARGS)
{
    int size = 0;
    Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
    size = headerSizeof(h, HEADER_MAGIC_NO);
    headerFree(h);
    PG_RETURN_INT32(size);
}

PG_FUNCTION_INFO_V1(header_gettagn);

Datum
header_gettagn(PG_FUNCTION_ARGS)
{
    Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
    int tag = PG_GETARG_INT32(1);
    int num = PG_GETARG_INT32(2);
    int_32 tagtype = 0;
    void *ret;
    int_32 count = 0;
    text * res = NULL;
    int size;
    if (headerGetEntry(h, tag, &tagtype, &ret, &count)) {
        switch(tagtype) {
            case RPM_STRING_TYPE:
                if(num == 0) { /* this kind of tag has only one entry, if another is want, no way */
                    size = strlen(ret) + VARHDRSZ;
                    res = palloc(size);
                    memcpy(VARDATA(res), ret, size - VARHDRSZ);
#ifdef SET_VARSIZE
                    SET_VARSIZE(res,size);
#else
                    VARATT_SIZEP(res) = size;
#endif
                }
                break;
            case RPM_STRING_ARRAY_TYPE:
                if (num >= 0 && num <= count) {
                    char ** sptr;
                    sptr = (char **)ret;
                    size = strlen(sptr[num]) + VARHDRSZ;
                    res = palloc(size);
                    memcpy(VARDATA(res), sptr[num], size - VARHDRSZ);
#ifdef SET_VARSIZE
                    SET_VARSIZE(res, size);
#else
                    VARATT_SIZEP(res) = size;
#endif
                }
                break;
            case RPM_INT8_TYPE:
            case RPM_INT16_TYPE:
            case RPM_INT32_TYPE:
                if (num >= 0 && num <= count) {
                    int *r;
                    char * fmt = palloc(16);
                    r = (int *)ret;
                    snprintf(fmt, 16, "%d", r[num]);
                    size = strlen(fmt) + VARHDRSZ;
                    res = palloc(size);
                    memcpy(VARDATA(res), fmt, size - VARHDRSZ);
#ifdef SET_VARSIZE
                    SET_VARSIZE(res, size);
#else
                    VARATT_SIZEP(res) = size;
#endif
                }
                break;
            case RPM_BIN_TYPE:
                if(num == 0) { /* same idea than for string */
                    size = count * sizeof(char) + VARHDRSZ;
                    res = palloc(size);
                    memcpy(VARDATA(res), ret, count);
#ifdef SET_VARSIZE
                    SET_VARSIZE(res, size);
#else
                    VARATT_SIZEP(res) = size;
#endif
                }
                break;
            default:
                /* Nothing */
                break;
        }
    }
    
    headerFreeData(ret, tagtype);
    headerFree(h);
    
    if(res)
        PG_RETURN_TEXT_P(res);
    else
        PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(header_query);

Datum
header_query(PG_FUNCTION_ARGS)
{
    FuncCallContext     *funcctx;
    int                  call_cntr;
    int                  max_calls;
    if (SRF_IS_FIRSTCALL()) {
        MemoryContext   oldcontext;
        Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
        int tag = PG_GETARG_INT32(1);
        int_32 tagtype = 0;
        void *ret;
        Datum **fmtret = NULL;
        int_32 count = 0;
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        if (headerGetEntry(h, tag, &tagtype, &ret, &count)) {
            switch(tagtype) {
                case RPM_STRING_TYPE:
                {
                    int size;
                    fmtret = (Datum **) palloc(sizeof(Datum *));
                    size = strlen(ret) + VARHDRSZ;
                    fmtret[0] = palloc(size);
                    memcpy(VARDATA(fmtret[0]), ret, size - VARHDRSZ);
#ifdef SET_VARSIZE
                    SET_VARSIZE(fmtret[0], size);
#else
                    VARATT_SIZEP(fmtret[0]) = size;
#endif
                    funcctx->max_calls = 1;
                }
                break;
                case RPM_STRING_ARRAY_TYPE:
                {
                    int i;
                    char ** sptr;
                    fmtret = (Datum **) palloc(count * sizeof(Datum *));
                    sptr = (char **)ret;
                    for (i = 0; i < count; i ++) {
                        int size = strlen(sptr[i]) + VARHDRSZ;
                        fmtret[i] = palloc(size);
                        memcpy(VARDATA(fmtret[i]), sptr[i], size - VARHDRSZ);
#ifdef SET_VARSIZE
                        SET_VARSIZE(fmtret[i], size);
#else
                        VARATT_SIZEP(fmtret[i]) = size;
#endif
                    }
                    funcctx->max_calls = count;
                }
                break;
                case RPM_INT8_TYPE:
                case RPM_INT16_TYPE:
                case RPM_INT32_TYPE:
                {
                    int i;
                    int *r;
                    fmtret = (Datum **) palloc(count * sizeof(Datum *));
                    r = (int *)ret;
                    for (i = 0; i < count; i ++) {
                        char * fmt = palloc(16);
                        int size;
                        snprintf(fmt, 16, "%d", r[i]);
                        size = strlen(fmt) + VARHDRSZ;
                        fmtret[i] = palloc(size);
                        memcpy(VARDATA(fmtret[i]), fmt, size - VARHDRSZ);
#ifdef SET_VARSIZE
                        SET_VARSIZE(fmtret[i], size);
#else
                        VARATT_SIZEP(fmtret[i]) = size;
#endif
                        pfree(fmt);
                    }
                    funcctx->max_calls = count;
                }
                break;
                case RPM_BIN_TYPE:
                {
                    fmtret = (Datum **) palloc(sizeof(Datum *));
                    fmtret[0] = palloc(count * sizeof(char) + VARHDRSZ);
                    memcpy(VARDATA(fmtret[0]), ret, count);
#ifdef SET_VARSIZE
                    SET_VARSIZE(fmtret[0], count + VARHDRSZ);
#else
                    VARATT_SIZEP(fmtret[0]) = count + VARHDRSZ;
#endif
                    funcctx->max_calls = 1;
                }
                break;
                default:
                break;
            }
            funcctx->user_fctx = fmtret;
        }
	headerFreeData(ret, tagtype);
        headerFree(h);
        MemoryContextSwitchTo(oldcontext);
    }
    funcctx = SRF_PERCALL_SETUP();
    
    call_cntr = funcctx->call_cntr;
    max_calls = funcctx->max_calls;
    if (call_cntr < max_calls)
    {
        Datum       **values = funcctx->user_fctx;

        SRF_RETURN_NEXT(funcctx, (Datum) values[call_cntr]);

    }
    else    /* do when there is no more left */
    {
        SRF_RETURN_DONE(funcctx);
    }
}

PG_FUNCTION_INFO_V1(header_files_query);

static inline void
string_palloc_snprintf_int(char **string, int value)
{
    *string = palloc(33);
    snprintf(*string, 32, "%d", value);
}

static inline void
string_palloc_memcpy(char **string, const char *value)
{
    if (value) {
        *string = palloc(strlen(value) + 1);
        memcpy(*string, value, strlen(value) + 1);
    } else {
        *string = palloc(1);
        *string[0] = '\0';
    }
}

Datum
header_files_query(PG_FUNCTION_ARGS)
{
    FuncCallContext     *funcctx;
    
    if (SRF_IS_FIRSTCALL()) {
        MemoryContext   oldcontext;
        Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
        rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, SCAREMEM);
        
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        funcctx->attinmeta = TupleDescGetAttInMetadata(
                RelationNameGetTupleDesc("_rpmfiles"));
        
        if (fi != NULL) {

        rpmfiInit(fi, 0);
        funcctx->max_calls = rpmfiFC(fi);
        funcctx->user_fctx = fi; /* keep track of data, extraction will done later */

        } else {

        funcctx->max_calls = 0;

        }
        
        h = headerFree(h);
    }
    funcctx = SRF_PERCALL_SETUP();
    
    if (funcctx->call_cntr < funcctx->max_calls)
    {
#define NB_COL 17
        HeapTuple    tuple;
        Datum        result;
        char ** values;
        int i = 0;
        rpmfi fi = funcctx->user_fctx;
        
        rpmfiNext(fi);

        values = (char **) palloc(NB_COL * sizeof(char *));
        
	    string_palloc_snprintf_int(&values[i++], rpmfiFX(fi) + 1);
        string_palloc_memcpy(&values[i++], rpmfiFN(fi));
        string_palloc_memcpy(&values[i++], rpmfiDN(fi));
        string_palloc_memcpy(&values[i++], rpmfiBN(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFFlags(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiVFlags(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFMode(fi));

        {
            const unsigned char * md5;
            if ((md5 = 
#if RPMMAJOR >= 4 && RPMVER >= 4 && RPMMINOR <= 4
                        rpmfiMD5(fi)
#else
                        rpmfiDigest(fi, NULL, NULL)
#endif
                        ) != NULL && *md5 != 0) {
                values[i] = palloc(33);
                (void) pgpHexCvt(values[i], md5, 16);
                values[i][32] = '\0';
            } else {
                values[i] = NULL;
            }
        }
        i++;
        
        {
            const char * link;
            if ((link = rpmfiFLink(fi)) != NULL && *link != 0) {
                string_palloc_memcpy(&values[i], link);
            } else {
                values[i] = NULL;
            }
        }
        i++;

        string_palloc_memcpy(&values[i++], rpmfiFUser(fi));
    	string_palloc_memcpy(&values[i++], rpmfiFGroup(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFInode(fi));
    	string_palloc_snprintf_int(&values[i++], rpmfiFSize(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFRdev(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFColor(fi));
        
        {
            const char * class;
            if ((class = rpmfiFClass(fi)) != NULL) {
                string_palloc_memcpy(&values[i], class);
            } else {
                values[i] = NULL;
            }
        }
        i++;

    	string_palloc_snprintf_int(&values[i++], rpmfiFMtime(fi));
	    string_palloc_snprintf_int(&values[i++], rpmfiFNlink(fi));

        tuple = BuildTupleFromCStrings(funcctx->attinmeta, values);
        result = HeapTupleGetDatum(tuple);

        for (i = 0; i < NB_COL; i++)
            if (values[i])
                pfree(values[i]);
        pfree(values);
#undef NB_COL
        SRF_RETURN_NEXT(funcctx, result);

    }
    else    /* do when there is no more left */
    {
        funcctx->user_fctx = rpmfiFree(funcctx->user_fctx);
        SRF_RETURN_DONE(funcctx);
    }
}

PG_FUNCTION_INFO_V1(header_deps_query);

Datum
header_deps_query(PG_FUNCTION_ARGS)
{
    FuncCallContext     *funcctx;
    
    if (SRF_IS_FIRSTCALL()) {
        MemoryContext   oldcontext;
        Header h = Datumgetheader(PG_GETARG_BYTEA_P_COPY(0), NULL);
        text * ttype = PG_GETARG_TEXT_P(1);
        char * type = VARDATA(ttype);
#ifdef DYING
        rpmds ds = NULL;
#endif

        /* w/o rpmds */
        rpmTag deptag = 0;
        rpmTag depFtag = 0;
        rpmTag depEVRtag = 0;
        pgds_t pgds;

        switch (*type) {
            case 'p':
            case 'P':
                deptag = RPMTAG_PROVIDENAME;
                depFtag = RPMTAG_PROVIDEFLAGS;
                depEVRtag = RPMTAG_PROVIDEVERSION;
                break;
            case 'r':
            case 'R':
                deptag = RPMTAG_REQUIRENAME;
                depFtag = RPMTAG_REQUIREFLAGS;
                depEVRtag = RPMTAG_REQUIREVERSION;
                break;
            case 'o':
            case 'O':
                deptag = RPMTAG_OBSOLETENAME;
                depFtag = RPMTAG_OBSOLETEFLAGS;
                depEVRtag = RPMTAG_OBSOLETEVERSION;
                break;
            case 'c':
            case 'C':
                deptag = RPMTAG_CONFLICTNAME;
                depFtag = RPMTAG_CONFLICTFLAGS;
                depEVRtag = RPMTAG_CONFLICTVERSION;
                break;
            case 't':
            case 'T':
                deptag = RPMTAG_TRIGGERNAME;
                depFtag = RPMTAG_TRIGGERFLAGS;
                depEVRtag = RPMTAG_TRIGGERVERSION;
                break;
            case 'e':
            case 'E':
                deptag = RPMTAG_ENHANCESNAME;
                depFtag = RPMTAG_ENHANCESFLAGS;
                depEVRtag = RPMTAG_ENHANCESVERSION;
                break;
            case 's':
            case 'S':
                deptag = RPMTAG_SUGGESTSNAME;
                depFtag = RPMTAG_SUGGESTSFLAGS;
                depEVRtag = RPMTAG_SUGGESTSVERSION;
                break;
            default:
                ereport(ERROR, /* TODO */
                    (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                     errmsg("Don't know this kind of dependencies")));
                break;
        }
        
        funcctx = SRF_FIRSTCALL_INIT();
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
        funcctx->attinmeta = TupleDescGetAttInMetadata(
                RelationNameGetTupleDesc("_rpmdeps"));

        pgds = palloc(sizeof(*pgds));
        memset(pgds, 0, sizeof(*pgds));
        funcctx->user_fctx = pgds;
        pgds->tag = deptag;
        pgds->h = h;
        (void) headerGetEntry(h, deptag, &pgds->Nametype,
                (void **) &pgds->Name, &pgds->Namecount);
        (void) headerGetEntry(h, depFtag, &pgds->Flagstype, 
                (void **) &pgds->Flags, &pgds->Flagscount);
        (void) headerGetEntry(h, depEVRtag, &pgds->EVRtype, 
                (void **) &pgds->EVR, &pgds->EVRcount);
        funcctx->max_calls = pgds->Namecount;
    }
    funcctx = SRF_PERCALL_SETUP();
    
    if (funcctx->call_cntr < funcctx->max_calls)
    {
#define NB_COL 6
        HeapTuple    tuple;
        Datum        result;
        char ** values;
        int i = 0;

        pgds_t pgds = funcctx->user_fctx;
        values = (char **) palloc(NB_COL * sizeof(char *));

	    string_palloc_snprintf_int(&values[i++], funcctx->call_cntr + 1);
	    string_palloc_snprintf_int(&values[i++], pgds->tag);
	    string_palloc_memcpy(&values[i++], pgds->Name[funcctx->call_cntr]);
    	string_palloc_snprintf_int(&values[i++], 
                funcctx->call_cntr < pgds->Flagscount
                  ? pgds->Flags[funcctx->call_cntr]
                  : 0);
	    string_palloc_memcpy(&values[i++], 
                funcctx->call_cntr < pgds->EVRcount
                  ? pgds->EVR[funcctx->call_cntr]
                  : "");
    	string_palloc_snprintf_int(&values[i++], /* rpmdsColor(ds) */ 0);

        tuple = BuildTupleFromCStrings(funcctx->attinmeta, values);
        result = HeapTupleGetDatum(tuple);

        for (i = 0; i < NB_COL; i++)
            if (values[i])
                pfree(values[i]);
        pfree(values);
#undef NB_COL
        SRF_RETURN_NEXT(funcctx, result);

    }
    else    /* do when there is no more left */
    {
        pgds_t pgds = funcctx->user_fctx;
        headerFreeData(pgds->Name, pgds->Nametype);
        headerFreeData(pgds->EVR, pgds->EVRtype);
        headerFreeData(pgds->Flags, pgds->Flagstype);
        pgds->h = headerFree(pgds->h);
        pfree(pgds); pgds = NULL;
        
        SRF_RETURN_DONE(funcctx);
    }
}

